from typing import Iterable, List, Optional

from .auxiliary import AuxiliaryDecrypt, AuxiliaryEncrypt
from .data_store import DataStore
from .guardian import Guardian
from .key_ceremony import (
    AuxiliaryPublicKey,
    CeremonyDetails,
    ElectionJointKey,
    ElectionPartialKeyBackup,
    ElectionPartialKeyChallenge,
    ElectionPartialKeyVerification,
    ElectionPublicKey,
    GuardianPair,
    PublicKeySet,
    combine_election_public_keys,
)
from .logs import log_warning
from .rsa import rsa_decrypt, rsa_encrypt
from .types import GUARDIAN_ID


class KeyCeremonyMediator:
    """
    KeyCeremonyMediator for assisting communication between guardians 
    """

    ceremony_details: CeremonyDetails

    _auxiliary_public_keys: DataStore[GUARDIAN_ID, AuxiliaryPublicKey]
    _election_public_keys: DataStore[GUARDIAN_ID, ElectionPublicKey]
    _election_partial_key_backups: DataStore[GuardianPair, ElectionPartialKeyBackup]
    _election_partial_key_challenges: DataStore[
        GuardianPair, ElectionPartialKeyChallenge
    ]
    _election_partial_key_verifications: DataStore[
        GuardianPair, ElectionPartialKeyVerification
    ]
    _guardians: List[Guardian] = []

    def __init__(self, ceremony_details: CeremonyDetails):
        self.ceremony_details = ceremony_details
        self._auxiliary_public_keys = DataStore[GUARDIAN_ID, AuxiliaryPublicKey]()
        self._election_public_keys = DataStore[GUARDIAN_ID, ElectionPublicKey]()
        self._election_partial_key_backups = DataStore[
            GuardianPair, ElectionPartialKeyBackup
        ]()
        self._election_partial_key_verifications = DataStore[
            GuardianPair, ElectionPartialKeyVerification
        ]()
        self._election_partial_key_challenges = DataStore[
            GuardianPair, ElectionPartialKeyChallenge
        ]()

    def announce(self, guardian: Guardian) -> None:
        """
        Announce the guardian as present and participating the Key Ceremony
        :param guardian: The guardian to announce
        """
        self.confirm_presence_of_guardian(guardian.share_public_keys())
        self._guardians.append(guardian)

        # When all guardians have announced, share the public keys among them
        if self.all_guardians_in_attendance():
            for sender in self._guardians:
                for recipient in self._guardians:
                    if sender.object_id != recipient.object_id:
                        recipient.save_guardian_public_keys(sender.share_public_keys())

    def orchestrate(
        self, encrypt: AuxiliaryEncrypt = rsa_encrypt
    ) -> Optional[List[Guardian]]:
        """
        Orchestrate the KLey Ceremony by sharing keys among the anounced guardians

        :param encrypt: Auxiliary encrypt function
        :return: a collection of guardians, or None if there is an error
        """
        if not self.all_guardians_in_attendance():
            return None

        # Partial Key Backup Generation
        for guardian in self._guardians:
            guardian.generate_election_partial_key_backups(encrypt)

        # Share Partial Key Backup
        for sender in self._guardians:
            for recipient in self._guardians:
                if sender.object_id != recipient.object_id:
                    backup = sender.share_election_partial_key_backup(
                        recipient.object_id
                    )

                    if backup is not None:
                        self.receive_election_partial_key_backup(backup)
                    else:
                        log_warning(
                            f"orchestrate failed sender {sender.object_id} could not share backup with recipient: {recipient.object_id}"
                        )
                        return None

        # Save the backups
        if self.all_election_partial_key_backups_available:
            for recipient_guardian in self._guardians:
                backups = self.share_election_partial_key_backups_to_guardian(
                    recipient_guardian.object_id
                )
                for backup in backups:
                    recipient_guardian.save_election_partial_key_backup(backup)

        return self._guardians

    def verify(self, decrypt: AuxiliaryDecrypt = rsa_decrypt) -> bool:
        """
        Verify that the guardians correctly shared keys
        :param decrypt: Auxiliary decrypt function
        :return: True if verification succeds, else False
        """
        for recipient in self._guardians:
            for sender in self._guardians:
                if sender.object_id != recipient.object_id:
                    verification = recipient.verify_election_partial_key_backup(
                        sender.object_id, decrypt
                    )
                    if verification is not None:
                        self.receive_election_partial_key_verification(verification)
                    else:
                        log_warning(
                            f"verify failed recipient {recipient.object_id} could not verify backup from sender: {sender.object_id}"
                        )
                        return False

        return (
            self.all_election_partial_key_verifications_received()
            and self.all_election_partial_key_backups_verified()
        )

    def reset(self, ceremony_details: CeremonyDetails) -> None:
        """
        Reset mediator to initial state
        :param ceremony_details: Ceremony details of election
        """
        self.ceremony_details = ceremony_details
        self._auxiliary_public_keys.clear()
        self._election_public_keys.clear()
        self._election_partial_key_backups.clear()
        self._election_partial_key_challenges.clear()
        self._election_partial_key_verifications.clear()
        self._guardians.clear()

    # Attendance
    def confirm_presence_of_guardian(self, public_key_set: PublicKeySet) -> None:
        """
        Confirm presence of guardian by passing their public key set
        :param public_key_set: Public key set
        """
        self.receive_auxiliary_public_key(
            AuxiliaryPublicKey(
                public_key_set.owner_id,
                public_key_set.sequence_order,
                public_key_set.auxiliary_public_key,
            )
        )
        self.receive_election_public_key(
            ElectionPublicKey(
                public_key_set.owner_id,
                public_key_set.election_public_key_proof,
                public_key_set.election_public_key,
            ),
        )

    def all_guardians_in_attendance(self) -> bool:
        """
        Check the attendance of all the guardians expected
        :return: True if all guardians in attendance
        """
        return (
            self.all_auxiliary_public_keys_available()
            and self.all_election_public_keys_available()
        )

    def share_guardians_in_attendance(self) -> Iterable[GUARDIAN_ID]:
        """
        Share a list of all the guardians in attendance
        :return: list of guardians ids
        """
        return self._election_public_keys.keys()

    # Auxiliary Public Keys
    def receive_auxiliary_public_key(self, public_key: AuxiliaryPublicKey) -> None:
        """
        Receive auxiliary public key from guardian
        :param public_key: Auxiliary public key
        """
        self._auxiliary_public_keys.set(public_key.owner_id, public_key)

    def all_auxiliary_public_keys_available(self) -> bool:
        """
        True if all auxiliary public key for all guardians available
        :return: All auxiliary public backups for all guardians available
        """
        return (
            len(self._auxiliary_public_keys)
            == self.ceremony_details.number_of_guardians
        )

    def share_auxiliary_public_keys(self) -> Iterable[AuxiliaryPublicKey]:
        """
        Share all currently stored auxiliary public keys for all guardians
        :return: list of auxiliary public keys
        """
        return self._auxiliary_public_keys.values()

    # Election Public Keys
    def receive_election_public_key(self, public_key: ElectionPublicKey) -> None:
        """
        Receive election public key from guardian
        :param public_key: election public key
        """
        self._election_public_keys.set(public_key.owner_id, public_key)

    def all_election_public_keys_available(self) -> bool:
        """
        True if all election public keys for all guardians available
        :return: All election public keys for all guardians available
        """
        return (
            len(self._election_public_keys) == self.ceremony_details.number_of_guardians
        )

    def share_election_public_keys(self) -> Iterable[ElectionPublicKey]:
        """
        Share all currently stored election public keys for all guardians
        :return: list of election public keys
        """
        return self._election_public_keys.values()

    # Election Partial Key Backups
    def receive_election_partial_key_backup(
        self, backup: ElectionPartialKeyBackup
    ) -> bool:
        """
        Receive election partial key backup from guardian
        :param backup: Election partial key backup
        :return: boolean indicating success or failure
        """
        if backup.owner_id == backup.designated_id:
            return False
        self._election_partial_key_backups.set(
            GuardianPair(backup.owner_id, backup.designated_id), backup
        )
        return True

    def all_election_partial_key_backups_available(self) -> bool:
        """
        True if all election partial key backups for all guardians available
        :return: All election partial key backups for all guardians available
        """
        required_backups_per_guardian = self.ceremony_details.number_of_guardians - 1
        return (
            len(self._election_partial_key_backups)
            == required_backups_per_guardian * self.ceremony_details.number_of_guardians
        )

    def share_election_partial_key_backups_to_guardian(
        self, guardian_id: GUARDIAN_ID
    ) -> List[ElectionPartialKeyBackup]:
        """
        Share all election partial key backups for designated guardian
        :param guardian_id: Recipients guardian id
        :return: List of guardians designated backups
        """
        backups: List[ElectionPartialKeyBackup] = []
        for current_guardian_id in self.share_guardians_in_attendance():
            if guardian_id != current_guardian_id:
                backup = self._election_partial_key_backups.get(
                    GuardianPair(current_guardian_id, guardian_id)
                )
                if backup is not None:
                    backups.append(backup)
        return backups

    # Partial Key Verifications
    def receive_election_partial_key_verification(
        self, verification: ElectionPartialKeyVerification
    ) -> None:
        """
        Receive election partial key verification from guardian
        :param verification: Election partial key verification
        """
        if verification.owner_id == verification.designated_id:
            return
        self._election_partial_key_verifications.set(
            GuardianPair(verification.owner_id, verification.designated_id),
            verification,
        )

    def all_election_partial_key_verifications_received(self) -> bool:
        """
        True if all election partial key verifications recieved
        :return: All election partial key verifications received
        """
        required_verifications_per_guardian = (
            self.ceremony_details.number_of_guardians - 1
        )
        return (
            len(self._election_partial_key_verifications)
            == required_verifications_per_guardian
            * self.ceremony_details.number_of_guardians
        )

    def all_election_partial_key_backups_verified(self) -> bool:
        """
        True if all election partial key backups verified
        :return: All election partial key backups verified
        """
        if not self.all_election_partial_key_verifications_received():
            return False
        for verification in self._election_partial_key_verifications.values():
            if not verification.verified:
                return False
        return True

    # Partial Key Challenges
    def share_failed_partial_key_verifications(self) -> List[GuardianPair]:
        """
        Share list of guardians with failed partial key backup verifications
        :return: List of guardian pairs with failed verifications
        """
        failed_verifications: List[GuardianPair] = []
        for pair, verification in self._election_partial_key_verifications.items():
            if not verification.verified:
                failed_verifications.append(pair)
        return failed_verifications

    def share_missing_election_partial_key_challenges(self) -> List[GuardianPair]:
        """
        Share list of guardians with missing election partial key challenges
        :return: List of guardian pairs with failed verifications and no challenges
        """
        failed_verifications = self.share_failed_partial_key_verifications()
        for pair in self._election_partial_key_challenges.keys():
            failed_verifications.remove(pair)
        return failed_verifications

    def receive_election_partial_key_challenge(
        self, challenge: ElectionPartialKeyChallenge
    ) -> None:
        """
        Receive an election partial key challenge from a guardian with a failed verification
        :param challenge: Election partial key challenge
        """
        self._election_partial_key_challenges.set(
            GuardianPair(challenge.owner_id, challenge.designated_id), challenge
        )

    def share_open_election_partial_key_challenges(
        self,
    ) -> List[ElectionPartialKeyChallenge]:
        """
        Share all open election partial key challenges with guardians
        :return: List of open election partial key challenges 
        """
        return list(self._election_partial_key_challenges.values())

    # Publish Joint Key
    def publish_joint_key(self) -> Optional[ElectionJointKey]:
        """
        Publish joint election key from the public keys of all guardians
        :return: Optional joint key for election
        """
        if not self.all_election_public_keys_available():
            return None
        if not self.all_election_partial_key_backups_verified():
            return None
        return combine_election_public_keys(self._election_public_keys)



from typing import Dict, Optional, Tuple

from .ballot import CiphertextAcceptedBallot, CiphertextSelection
from .decryption_share import (
    BallotDecryptionShare,
    CiphertextDecryptionSelection,
    TallyDecryptionShare,
    get_ballot_shares_for_selection,
    get_tally_shares_for_selection,
)
from .dlog import discrete_log
from .election import CiphertextElectionContext
from .group import ElementModP, ElementModQ, mult_p, div_p
from .tally import (
    CiphertextTally,
    PlaintextTally,
    CiphertextTallyContest,
    PlaintextTallyContest,
    PlaintextTallySelection,
)
from .logs import log_warning
from .types import BALLOT_ID, CONTEST_ID, GUARDIAN_ID, SELECTION_ID

AVAILABLE_GUARDIAN_ID = GUARDIAN_ID
MISSING_GUARDIAN_ID = GUARDIAN_ID

GUARDIAN_PUBLIC_KEY = ElementModP

ELECTION_PUBLIC_KEY = ElementModP

# The methods in this file can be used to decrypt values if private keys or nonces are not known
# and the key ceremony is used to share secrets among a quorum of guardians


def decrypt_selection_with_decryption_shares(
    selection: CiphertextSelection,
    shares: Dict[
        GUARDIAN_ID, Tuple[ELECTION_PUBLIC_KEY, CiphertextDecryptionSelection]
    ],
    extended_base_hash: ElementModQ,
    suppress_validity_check: bool = False,
) -> Optional[PlaintextTallySelection]:
    """
    Decrypt the specified `CiphertextTallySelection` with the collection of `ElementModP` decryption shares.
    Each share is expected to be passed with the corresponding public key so that the encryption can be validated

    :param selection: a `CiphertextTallySelection`
    :param shares: the collection of shares to decrypt the selection
    :param extended_base_hash: the extended base hash code (𝑄') for the election
    :return: a `PlaintextTallySelection` or `None` if there is an error
    :param suppress_validity_check: do not validate the encryption prior to decrypting (useful for tests)
    """

    if not suppress_validity_check:
        # Verify that all of the shares are computed correctly
        for guardian_id, share in shares.items():
            public_key, decryption = share
            # verify we have a proof or recovered parts
            if not decryption.is_valid(
                selection.ciphertext, public_key, extended_base_hash
            ):
                return None

    # accumulate all of the shares calculated for the selection
    all_shares_product_M = mult_p(
        *[decryption.share for (_, decryption) in shares.values()]
    )

    # Calculate 𝑀=𝐵⁄(∏𝑀𝑖) mod 𝑝.
    decrypted_value = div_p(selection.ciphertext.data, all_shares_product_M)
    d_log = discrete_log(decrypted_value)
    return PlaintextTallySelection(
        selection.object_id,
        d_log,
        decrypted_value,
        selection.ciphertext,
        [share for (guardian_id, (public_key, share)) in shares.items()],
    )


def decrypt_tally_contests_with_decryption_shares(
    tally: Dict[CONTEST_ID, CiphertextTallyContest],
    shares: Dict[GUARDIAN_ID, TallyDecryptionShare],
    extended_base_hash: ElementModQ,
) -> Optional[Dict[CONTEST_ID, PlaintextTallyContest]]:
    """
    Decrypt the specified tally within the context of the specified Decryption Shares

    :param tally: the encrypted tally of contests
    :param shares: a collection of `TallyDecryptionShare` used to decrypt
    :param extended_base_hash: the extended base hash code (𝑄') for the election
    :return: a collection of `PlaintextTallyContest` or `None` if there is an error
    """
    contests: Dict[CONTEST_ID, PlaintextTallyContest] = {}

    # iterate through the tally contests
    for contest in tally.values():
        selections: Dict[SELECTION_ID, PlaintextTallySelection] = {}

        for selection in contest.tally_selections.values():
            tally_shares = get_tally_shares_for_selection(selection.object_id, shares)
            plaintext_selection = decrypt_selection_with_decryption_shares(
                selection, tally_shares, extended_base_hash
            )
            if plaintext_selection is None:
                log_warning(f"could not decrypt tally for contest {contest.object_id}")
                return None
            selections[plaintext_selection.object_id] = plaintext_selection

        contests[contest.object_id] = PlaintextTallyContest(
            contest.object_id, selections
        )

    return contests


def decrypt_tally(
    tally: CiphertextTally,
    shares: Dict[GUARDIAN_ID, TallyDecryptionShare],
    context: CiphertextElectionContext,
) -> Optional[PlaintextTally]:
    """
    Try to decrypt the tally and the spoiled ballots using the provided decryption shares
    :param tally: The CiphertextTally to decrypt
    :param shares: The guardian Decryption Shares for all guardians
    :param context: the Ciphertextelectioncontext
    :return: A PlaintextTally or None if there is an error
    """
    contests = decrypt_tally_contests_with_decryption_shares(
        tally.cast, shares, context.crypto_extended_base_hash
    )

    if contests is None:
        return None

    spoiled_ballots = decrypt_spoiled_ballots(
        tally.spoiled_ballots, shares, context.crypto_extended_base_hash
    )

    if spoiled_ballots is None:
        return None

    return PlaintextTally(tally.object_id, contests, spoiled_ballots)


def decrypt_spoiled_ballots(
    spoiled_ballots: Dict[BALLOT_ID, CiphertextAcceptedBallot],
    shares: Dict[AVAILABLE_GUARDIAN_ID, TallyDecryptionShare],
    extended_base_hash: ElementModQ,
) -> Optional[Dict[BALLOT_ID, Dict[CONTEST_ID, PlaintextTallyContest]]]:
    """
    Try to decrypt each of the spoiled ballots using the provided decryption shares
    """

    plaintext_spoiled_ballots: Dict[
        BALLOT_ID, Dict[CONTEST_ID, PlaintextTallyContest]
    ] = {}

    for spoiled_ballot in spoiled_ballots.values():
        ballot_shares: Dict[AVAILABLE_GUARDIAN_ID, BallotDecryptionShare] = {
            guardian_id: share.spoiled_ballots[spoiled_ballot.object_id]
            for guardian_id, share in shares.items()
        }

        decrypted_ballot = decrypt_ballot(
            spoiled_ballot, ballot_shares, extended_base_hash
        )
        if decrypted_ballot:
            plaintext_spoiled_ballots[spoiled_ballot.object_id] = decrypted_ballot
        else:
            return None

    return plaintext_spoiled_ballots


def decrypt_ballot(
    ballot: CiphertextAcceptedBallot,
    shares: Dict[AVAILABLE_GUARDIAN_ID, BallotDecryptionShare],
    extended_base_hash: ElementModQ,
) -> Optional[Dict[CONTEST_ID, PlaintextTallyContest]]:
    """
    Try to decrypt a single ballot using the provided decryption shares
    """

    contests: Dict[CONTEST_ID, PlaintextTallyContest] = {}
    for contest in ballot.contests:
        selections: Dict[SELECTION_ID, PlaintextTallySelection] = {}
        for selection in contest.ballot_selections:
            selection_shares = get_ballot_shares_for_selection(
                selection.object_id, shares
            )
            plaintext_selection = decrypt_selection_with_decryption_shares(
                selection, selection_shares, extended_base_hash
            )

            # verify the plaintext values are received and add them to the collection
            if plaintext_selection is None:
                log_warning(
                    f"could not decrypt ballot {ballot.object_id} for contest {contest.object_id} selection {selection.object_id}"
                )
                return None
            selections[plaintext_selection.object_id] = plaintext_selection

        contests[contest.object_id] = PlaintextTallyContest(
            contest.object_id, selections
        )

    return contests



from __future__ import annotations
from contextlib import AbstractContextManager
from multiprocessing import Pool as ProcessPool
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing.pool import Pool
from psutil import cpu_count
import traceback
from typing import Any, Callable, Iterable, List

from .logs import log_warning
from .singleton import Singleton
from .utils import T


class Scheduler(Singleton, AbstractContextManager):
    """
    Worker that wraps Multprocessing and allows 
    for shared context or spawning processes.
    Implemented as a singleton to guarantee there is only one set
    of tread and process pools in use throughout the library.
    Also implements the [Context Manager Protocol](https://docs.python.org/3.8/library/stdtypes.html#typecontextmanager)
    """

    __process_pool: Pool
    __thread_pool: Pool

    def __init__(self) -> None:
        super(Scheduler, self).__init__()
        self._open()

    def __enter__(self) -> Scheduler:
        self._open()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        self.close()

    def _open(self) -> None:
        """Open pools"""
        max_processes = cpu_count(logical=False)
        # Reserve one CPU for I/O bound tasks
        if max_processes > 2:
            max_processes = max_processes - 1
        self.__process_pool = ProcessPool(max_processes)
        self.__thread_pool = ThreadPool(max_processes)

    def close(self) -> None:
        """Close pools"""
        self.__process_pool.close()
        self.__thread_pool.close()

    def cpu_count(self) -> int:
        """Get CPU count"""
        return int(cpu_count(logical=False))

    def schedule(
        self,
        task: Callable,
        arguments: Iterable[Iterable[Any]],
        with_shared_resources: bool = False,
    ) -> List[T]:
        """
        Schedule tasks with list of arguments
        :param task: the callable task to execute
        :param arguments: the list of lists passed to the task using starmap
        :param with_shared_resources: flag to use threads instead of processes
                                      allowing resources to be shared.  note
                                      when using the threadpool, execution is bound
                                      by default to the [global interpreter lock](https://docs.python.org/3.8/glossary.html#term-global-interpreter-lock)
        """
        if with_shared_resources:
            return self.safe_starmap(self.__thread_pool, task, arguments)
        return self.safe_starmap(self.__process_pool, task, arguments)

    def safe_starmap(
        self, pool: Pool, task: Callable, arguments: Iterable[Iterable[Any]]
    ) -> List[T]:
        """Safe wrapper around starmap to ensure pool is open"""
        try:
            return pool.starmap(task, arguments)
        except ValueError as e:
            log_warning(
                f"safe_starmap({task}, {arguments}) exception ValueError({str(e)})"
            )
            return []
        except Exception:
            log_warning(
                f"safe_starmap({task}, {arguments}) failed with \n {traceback.format_exc()}"
            )
            return []

    def safe_map(self, pool: Pool, task: Callable, arguments: Iterable[Any]) -> List[T]:
        """Safe wrapper around starmap to ensure pool is open"""
        try:
            return pool.map(task, arguments)
        except ValueError as e:
            log_warning(f"safe_map({task}, {arguments}) exception ValueError({str(e)})")
            return []
        except Exception:
            log_warning(
                f"safe_starmap({task}, {arguments}) failed with \n {traceback.format_exc()}"
            )
            return []



from enum import Enum

from .serializable import Serializable
from .utils import space_between_capitals


class ProofUsage(Enum):
    """Usage case for proof"""

    Unknown = "Unknown"
    SecretValue = "Prove knowledge of secret value"
    SelectionLimit = "Prove value within selection's limit"
    SelectionValue = "Prove selection's value (0 or 1)"


class Proof(Serializable):
    """Base class for proofs with name and usage case"""

    name: str = "Proof"
    usage: ProofUsage = ProofUsage.Unknown

    def __init__(self) -> None:
        object.__setattr__(
            self, "name", space_between_capitals(self.__class__.__name__)
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Tuple

from .election import (
    CiphertextElectionContext,
    ElectionDescription,
    InternalElectionDescription,
    make_ciphertext_election_context,
)
from .group import ElementModP
from .utils import get_optional


@dataclass
class ElectionBuilder(object):
    """
    `ElectionBuilder` is a stateful builder object that constructs `CiphertextElectionContext` objects
    following the initialization process that ElectionGuard Expects.
    """

    number_of_guardians: int
    """
    The number of guardians necessary to generate the public key
    """
    quorum: int
    """
    The quorum of guardians necessary to decrypt an election.  Must be less than `number_of_guardians`
    """

    description: ElectionDescription

    internal_description: InternalElectionDescription = field(init=False)

    elgamal_public_key: Optional[ElementModP] = field(default=None)

    def __post_init__(self) -> None:
        self.internal_description = InternalElectionDescription(self.description)

    def set_public_key(self, elgamal_public_key: ElementModP) -> ElectionBuilder:
        """
        Set election public key
        :param elgamal_public_key: elgamal public key for election
        :return: election builder
        """
        self.elgamal_public_key = elgamal_public_key
        return self

    def build(
        self,
    ) -> Optional[Tuple[InternalElectionDescription, CiphertextElectionContext]]:
        """
        Build election
        :return: election description and context or none
        """
        if not self.description.is_valid():
            return None

        if self.elgamal_public_key is None:
            return None

        return (
            self.internal_description,
            make_ciphertext_election_context(
                self.number_of_guardians,
                self.quorum,
                get_optional(self.elgamal_public_key),
                self.description.crypto_hash(),
            ),
        )



from typing import List, Dict
from electionguard.ballot import PlaintextBallot


def accumulate_plaintext_ballots(ballots: List[PlaintextBallot]) -> Dict[str, int]:
    """
    Internal helper function for testing: takes a list of plaintext ballots as input,
    digs into all of the individual selections and then accumulates them, using
    their `object_id` fields as keys. This function only knows what to do with
    `n_of_m` elections. It's not a general-purpose tallying mechanism for other
    election types.

    :param ballots: a list of plaintext ballots
    :return: a dict from selection object_id's to integer totals
    """
    tally: Dict[str, int] = {}
    for ballot in ballots:
        for contest in ballot.contests:
            for selection in contest.ballot_selections:
                assert (
                    not selection.is_placeholder_selection
                ), "Placeholder selections should not exist in the plaintext ballots"
                desc_id = selection.object_id
                if desc_id not in tally:
                    tally[desc_id] = 0
                # returns 1 or 0 for n-of-m ballot selections
                tally[desc_id] += selection.to_int()
    return tally



#!/usr/bin/env python
from random import randint
from shutil import rmtree
from typing import List
import uuid

from electionguardtest.election_factory import ElectionFactory, QUORUM
from electionguardtest.ballot_factory import BallotFactory

from electionguard.ballot import (
    CiphertextBallot,
    CiphertextAcceptedBallot,
)
from electionguard.ballot_store import BallotStore
from electionguard.ballot_box import BallotBox
from electionguard.decryption_mediator import DecryptionMediator
from electionguard.encrypt import EncryptionDevice, EncryptionMediator
from electionguard.publish import publish, publish_private_data, RESULTS_DIR
from electionguard.tally import tally_ballots, publish_ciphertext_tally
from electionguard.utils import get_optional

DEFAULT_NUMBER_OF_BALLOTS = 5
DEFAULT_SPOIL_RATE = 50
DEFAULT_USE_ALL_GUARDIANS = False


class ElectionSampleDataGenerator:
    """
    Generates sample data for an example election using the "Hamilton County" data set.
    """

    election_factory: ElectionFactory
    ballot_factory: BallotFactory

    encryption_device: EncryptionDevice
    encrypter: EncryptionMediator

    def __init__(self) -> None:
        """Initialize the class"""
        self.election_factory = ElectionFactory()
        self.ballot_factory = BallotFactory()
        self.encryption_device = EncryptionDevice(f"polling-place-{str(uuid.uuid1())}")

    def generate(
        self,
        number_of_ballots: int = DEFAULT_NUMBER_OF_BALLOTS,
        spoil_rate: int = DEFAULT_SPOIL_RATE,
        use_all_guardians: bool = DEFAULT_USE_ALL_GUARDIANS,
    ):
        """
        Generate the sample data set
        """

        # Clear the results directory
        rmtree(RESULTS_DIR, ignore_errors=True)

        # Configure the election
        (
            public_data,
            private_data,
        ) = self.election_factory.get_hamilton_election_with_encryption_context()
        plaintext_ballots = self.ballot_factory.generate_fake_plaintext_ballots_for_election(
            public_data.metadata, number_of_ballots
        )
        self.encrypter = EncryptionMediator(
            public_data.metadata, public_data.context, self.encryption_device
        )

        # Encrypt some ballots
        ciphertext_ballots: List[CiphertextBallot] = []
        for plaintext_ballot in plaintext_ballots:
            ciphertext_ballots.append(
                get_optional(self.encrypter.encrypt(plaintext_ballot))
            )

        ballot_store = BallotStore()
        ballot_box = BallotBox(public_data.metadata, public_data.context, ballot_store)

        # Randomly cast/spoil the ballots
        accepted_ballots: List[CiphertextAcceptedBallot] = []
        for ballot in ciphertext_ballots:
            if randint(0, 100) < spoil_rate:
                accepted_ballots.append(ballot_box.spoil(ballot))
            else:
                accepted_ballots.append(ballot_box.cast(ballot))

        # Tally
        ciphertext_tally = get_optional(
            tally_ballots(ballot_store, public_data.metadata, public_data.context)
        )

        # Decrypt
        decrypter = DecryptionMediator(
            public_data.metadata, public_data.context, ciphertext_tally
        )

        for i, guardian in enumerate(private_data.guardians):
            if use_all_guardians or i < QUORUM:
                decrypter.announce(guardian)

        plaintext_tally = get_optional(decrypter.get_plaintext_tally())

        # Publish
        publish(
            public_data.description,
            public_data.context,
            public_data.constants,
            [self.encryption_device],
            accepted_ballots,
            ciphertext_tally.spoiled_ballots.values(),
            publish_ciphertext_tally(ciphertext_tally),
            plaintext_tally,
            public_data.guardians,
        )

        publish_private_data(
            plaintext_ballots, ciphertext_ballots, private_data.guardians
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Generate sample ballot data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-n",
        "--number-of-ballots",
        metavar="<count>",
        default=DEFAULT_NUMBER_OF_BALLOTS,
        type=int,
        help="The number of ballots to generate.",
    )
    parser.add_argument(
        "-s",
        "--spoil-rate",
        metavar="<rate>",
        default=DEFAULT_SPOIL_RATE,
        type=int,
        help="The approximate percentage of total ballots to spoil instead of cast. Provide a number from 0-100.",
    )
    parser.add_argument(
        "-a",
        "--all-guardians",
        default=DEFAULT_USE_ALL_GUARDIANS,
        action="store_true",
        help="If specified, all guardians will be included.  Otherwise, only the threshold number will be included.",
    )
    args = parser.parse_args()

    ElectionSampleDataGenerator().generate(
        args.number_of_ballots, args.spoil_rate, args.all_guardians
    )



from typing import List, Optional
from uuid import getnode

from .ballot import (
    CiphertextBallot,
    CiphertextBallotContest,
    CiphertextBallotSelection,
    PlaintextBallot,
    PlaintextBallotContest,
    PlaintextBallotSelection,
    make_ciphertext_ballot_contest,
    make_ciphertext_ballot_selection,
    make_ciphertext_ballot,
)

from .election import (
    CiphertextElectionContext,
    InternalElectionDescription,
    ContestDescription,
    ContestDescriptionWithPlaceholders,
    SelectionDescription,
)
from .elgamal import elgamal_encrypt
from .group import ElementModP, ElementModQ, rand_q
from .logs import log_warning
from .nonces import Nonces
from .serializable import Serializable
from .tracker import get_hash_for_device
from .utils import get_optional, get_or_else_optional_func


class EncryptionDevice(Serializable):
    """
    Metadata for encryption device
    """

    uuid: int
    location: str

    def __init__(self, location: str) -> None:
        self.uuid = generate_device_uuid()
        self.location = location

    def get_hash(self) -> ElementModQ:
        """
        Get hash for encryption device
        :return: Starting hash
        """
        return get_hash_for_device(self.uuid, self.location)


class EncryptionMediator(object):
    """
    An object for caching election and encryption state.

    It composes Elections and Ballots.
    """

    _metadata: InternalElectionDescription
    _encryption: CiphertextElectionContext
    _seed_hash: ElementModQ

    def __init__(
        self,
        election_metadata: InternalElectionDescription,
        context: CiphertextElectionContext,
        encryption_device: EncryptionDevice,
    ):
        self._metadata = election_metadata
        self._encryption = context
        self._seed_hash = encryption_device.get_hash()

    def encrypt(self, ballot: PlaintextBallot) -> Optional[CiphertextBallot]:
        """
        Encrypt the specified ballot using the cached election context.
        """
        encrypted_ballot = encrypt_ballot(
            ballot, self._metadata, self._encryption, self._seed_hash
        )
        if encrypted_ballot is not None and encrypted_ballot.tracking_hash is not None:
            self._seed_hash = encrypted_ballot.tracking_hash
        return encrypted_ballot


def generate_device_uuid() -> int:
    """
    Get unique identifier for device
    :return: Unique identifier
    """
    return getnode()


def selection_from(
    description: SelectionDescription,
    is_placeholder: bool = False,
    is_affirmative: bool = False,
) -> PlaintextBallotSelection:
    """
    Construct a `BallotSelection` from a specific `SelectionDescription`.
    This function is useful for filling selections when a voter undervotes a ballot.
    It is also used to create placeholder representations when generating the `ConstantChaumPedersenProof`

    :param description: The `SelectionDescription` which provides the relevant `object_id`
    :param is_placeholder: Mark this selection as a placeholder value
    :param is_affirmative: Mark this selection as `yes`
    :return: A BallotSelection
    """

    return PlaintextBallotSelection(
        description.object_id,
        vote=str(is_affirmative),
        is_placeholder_selection=is_placeholder,
    )


def contest_from(description: ContestDescription) -> PlaintextBallotContest:
    """
    Construct a `BallotContest` from a specific `ContestDescription` with all false fields.
    This function is useful for filling contests and selections when a voter undervotes a ballot.

    :param description: The `ContestDescription` used to derive the well-formed `BallotContest`
    :return: a `BallotContest`
    """

    selections: List[PlaintextBallotSelection] = list()

    for selection_description in description.ballot_selections:
        selections.append(selection_from(selection_description))

    return PlaintextBallotContest(description.object_id, selections)


def encrypt_selection(
    selection: PlaintextBallotSelection,
    selection_description: SelectionDescription,
    elgamal_public_key: ElementModP,
    crypto_extended_base_hash: ElementModQ,
    nonce_seed: ElementModQ,
    is_placeholder: bool = False,
    should_verify_proofs: bool = True,
) -> Optional[CiphertextBallotSelection]:
    """
    Encrypt a specific `BallotSelection` in the context of a specific `BallotContest`

    :param selection: the selection in the valid input form
    :param selection_description: the `SelectionDescription` from the `ContestDescription` which defines this selection's structure
    :param elgamal_public_key: the public key (K) used to encrypt the ballot
    :param crypto_extended_base_hash: the extended base hash of the election
    :param nonce_seed: an `ElementModQ` used as a header to seed the `Nonce` generated for this selection.
                 this value can be (or derived from) the BallotContest nonce, but no relationship is required
    :param is_placeholder: specifies if this is a placeholder selection
    :param should_verify_proofs: specify if the proofs should be verified prior to returning (default True)
    """

    # Validate Input
    if not selection.is_valid(selection_description.object_id):
        log_warning(f"malformed input selection: {selection}")
        return None

    selection_description_hash = selection_description.crypto_hash()
    nonce_sequence = Nonces(selection_description_hash, nonce_seed)
    selection_nonce = nonce_sequence[selection_description.sequence_order]
    disjunctive_chaum_pedersen_nonce = next(iter(nonce_sequence))

    selection_representation = selection.to_int()

    # Generate the encryption
    elgamal_encryption = elgamal_encrypt(
        selection_representation, selection_nonce, elgamal_public_key
    )

    if elgamal_encryption is None:
        # will have logged about the failure earlier, so no need to log anything here
        return None

    # TODO: ISSUE #35: encrypt/decrypt: encrypt the extended_data field

    # Create the return object
    encrypted_selection = make_ciphertext_ballot_selection(
        object_id=selection.object_id,
        description_hash=selection_description_hash,
        ciphertext=get_optional(elgamal_encryption),
        elgamal_public_key=elgamal_public_key,
        crypto_extended_base_hash=crypto_extended_base_hash,
        proof_seed=disjunctive_chaum_pedersen_nonce,
        selection_representation=selection_representation,
        is_placeholder_selection=is_placeholder,
        nonce=selection_nonce,
    )

    if encrypted_selection.proof is None:
        return None  # log will have happened earlier

    # optionally, skip the verification step
    if not should_verify_proofs:
        return encrypted_selection

    # verify the selection.
    if encrypted_selection.is_valid_encryption(
        selection_description_hash, elgamal_public_key, crypto_extended_base_hash
    ):
        return encrypted_selection
    else:
        log_warning(
            f"mismatching selection proof for selection {encrypted_selection.object_id}"
        )
        return None


def encrypt_contest(
    contest: PlaintextBallotContest,
    contest_description: ContestDescriptionWithPlaceholders,
    elgamal_public_key: ElementModP,
    crypto_extended_base_hash: ElementModQ,
    nonce_seed: ElementModQ,
    should_verify_proofs: bool = True,
) -> Optional[CiphertextBallotContest]:
    """
    Encrypt a specific `BallotContest` in the context of a specific `Ballot`.

    This method accepts a contest representation that only includes `True` selections.
    It will fill missing selections for a contest with `False` values, and generate `placeholder`
    selections to represent the number of seats available for a given contest.  By adding `placeholder`
    votes

    :param contest: the contest in the valid input form
    :param contest_description: the `ContestDescriptionWithPlaceholders` from the `ContestDescription` which defines this contest's structure
    :param elgamal_public_key: the public key (k) used to encrypt the ballot
    :param crypto_extended_base_hash: the extended base hash of the election
    :param nonce_seed: an `ElementModQ` used as a header to seed the `Nonce` generated for this contest.
                 this value can be (or derived from) the Ballot nonce, but no relationship is required
    :param should_verify_proofs: specify if the proofs should be verified prior to returning (default True)
    """

    # Validate Input
    if not contest.is_valid(
        contest_description.object_id,
        len(contest_description.ballot_selections),
        contest_description.number_elected,
        contest_description.votes_allowed,
    ):
        log_warning(f"malformed input contest: {contest}")
        return None

    if not contest_description.is_valid():
        log_warning(f"malformed contest description: {contest_description}")
        return None

    # account for sequence id
    contest_description_hash = contest_description.crypto_hash()
    nonce_sequence = Nonces(contest_description_hash, nonce_seed)
    contest_nonce = nonce_sequence[contest_description.sequence_order]
    chaum_pedersen_nonce = next(iter(nonce_sequence))

    encrypted_selections: List[CiphertextBallotSelection] = list()

    selection_count = 0

    # TODO: ISSUE #54 this code could be inefficient if we had a contest
    # with a lot of choices, although the O(n^2) iteration here is small
    # compared to the huge cost of doing the cryptography.

    # Generate the encrypted selections
    for description in contest_description.ballot_selections:
        has_selection = False
        encrypted_selection = None

        # iterate over the actual selections for each contest description
        # and apply the selected value if it exists.  If it does not, an explicit
        # false is entered instead and the selection_count is not incremented
        # this allows consumers to only pass in the relevant selections made by a voter
        for selection in contest.ballot_selections:
            if selection.object_id == description.object_id:
                # track the selection count so we can append the
                # appropriate number of true placeholder votes
                has_selection = True
                selection_count += selection.to_int()
                encrypted_selection = encrypt_selection(
                    selection,
                    description,
                    elgamal_public_key,
                    crypto_extended_base_hash,
                    contest_nonce,
                )
                break

        if not has_selection:
            # No selection was made for this possible value
            # so we explicitly set it to false
            encrypted_selection = encrypt_selection(
                selection_from(description),
                description,
                elgamal_public_key,
                crypto_extended_base_hash,
                contest_nonce,
            )

        if encrypted_selection is None:
            return None  # log will have happened earlier
        encrypted_selections.append(get_optional(encrypted_selection))

    # Handle Placeholder selections
    # After we loop through all of the real selections on the ballot,
    # we loop through each placeholder value and determine if it should be filled in

    # Add a placeholder selection for each possible seat in the contest
    for placeholder in contest_description.placeholder_selections:
        # for undervotes, select the placeholder value as true for each available seat
        # note this pattern is used since DisjunctiveChaumPedersen expects a 0 or 1
        # so each seat can only have a maximum value of 1 in the current implementation
        select_placeholder = False
        if selection_count < contest_description.number_elected:
            select_placeholder = True
            selection_count += 1

        encrypted_selection = encrypt_selection(
            selection=selection_from(
                description=placeholder,
                is_placeholder=True,
                is_affirmative=select_placeholder,
            ),
            selection_description=placeholder,
            elgamal_public_key=elgamal_public_key,
            crypto_extended_base_hash=crypto_extended_base_hash,
            nonce_seed=contest_nonce,
            is_placeholder=True,
            should_verify_proofs=True,
        )
        if encrypted_selection is None:
            return None  # log will have happened earlier
        encrypted_selections.append(get_optional(encrypted_selection))

    # TODO: ISSUE #33: support other cases such as cumulative voting
    # (individual selections being an encryption of > 1)
    if (
        contest_description.votes_allowed is not None
        and selection_count < contest_description.votes_allowed
    ):
        log_warning(
            "mismatching selection count: only n-of-m style elections are currently supported"
        )

    # Create the return object
    encrypted_contest = make_ciphertext_ballot_contest(
        object_id=contest.object_id,
        description_hash=contest_description_hash,
        ballot_selections=encrypted_selections,
        elgamal_public_key=elgamal_public_key,
        crypto_extended_base_hash=crypto_extended_base_hash,
        proof_seed=chaum_pedersen_nonce,
        number_elected=contest_description.number_elected,
        nonce=contest_nonce,
    )

    if encrypted_contest is None or encrypted_contest.proof is None:
        return None  # log will have happened earlier

    if not should_verify_proofs:
        return encrypted_contest

    # Verify the proof
    if encrypted_contest.is_valid_encryption(
        contest_description_hash, elgamal_public_key, crypto_extended_base_hash
    ):
        return encrypted_contest
    else:
        log_warning(
            f"mismatching contest proof for contest {encrypted_contest.object_id}"
        )
        return None


# TODO: ISSUE #57: add the device hash to the function interface so it can be propagated with the ballot.
# also propagate the seed hash so that the ballot tracking id's can be regenerated
# by traversing the collection of ballots encrypted by a specific device


def encrypt_ballot(
    ballot: PlaintextBallot,
    election_metadata: InternalElectionDescription,
    context: CiphertextElectionContext,
    seed_hash: ElementModQ,
    nonce: Optional[ElementModQ] = None,
    should_verify_proofs: bool = True,
) -> Optional[CiphertextBallot]:
    """
    Encrypt a specific `Ballot` in the context of a specific `CiphertextElectionContext`.

    This method accepts a ballot representation that only includes `True` selections.
    It will fill missing selections for a contest with `False` values, and generate `placeholder`
    selections to represent the number of seats available for a given contest.

    This method also allows for ballots to exclude passing contests for which the voter made no selections.
    It will fill missing contests with `False` selections and generate `placeholder` selections that are marked `True`.

    :param ballot: the ballot in the valid input form
    :param election_metadata: the `InternalElectionDescription` which defines this ballot's structure
    :param context: all the cryptographic context for the election
    :param seed_hash: Hash from previous ballot or starting hash from device
    :param nonce: an optional `int` used to seed the `Nonce` generated for this contest
                 if this value is not provided, the secret generating mechanism of the OS provides its own
    :param should_verify_proofs: specify if the proofs should be verified prior to returning (default True)
    """

    # Determine the relevant range of contests for this ballot style
    style = election_metadata.get_ballot_style(ballot.ballot_style)

    # Validate Input
    if not ballot.is_valid(style.object_id):
        log_warning(f"malformed input ballot: {ballot}")
        return None

    # Generate a random master nonce to use for the contest and selection nonce's on the ballot
    random_master_nonce = get_or_else_optional_func(nonce, lambda: rand_q())

    # Include a representation of the election and the external Id in the nonce's used
    # to derive other nonce values on the ballot
    nonce_seed = CiphertextBallot.nonce_seed(
        election_metadata.description_hash, ballot.object_id, random_master_nonce,
    )

    encrypted_contests: List[CiphertextBallotContest] = list()

    # only iterate on contests for this specific ballot style
    for description in election_metadata.get_contests_for(ballot.ballot_style):
        use_contest = None
        for contest in ballot.contests:
            if contest.object_id == description.object_id:
                use_contest = contest
                break
        # no selections provided for the contest, so create a placeholder contest
        if not use_contest:
            use_contest = contest_from(description)

        encrypted_contest = encrypt_contest(
            use_contest,
            description,
            context.elgamal_public_key,
            context.crypto_extended_base_hash,
            nonce_seed,
        )

        if encrypted_contest is None:
            return None  # log will have happened earlier
        encrypted_contests.append(get_optional(encrypted_contest))

    # Create the return object
    encrypted_ballot = make_ciphertext_ballot(
        ballot.object_id,
        ballot.ballot_style,
        election_metadata.description_hash,
        seed_hash,
        encrypted_contests,
        random_master_nonce,
    )

    if not encrypted_ballot.tracking_hash:
        return None

    if not should_verify_proofs:
        return encrypted_ballot

    # Verify the proofs
    if encrypted_ballot.is_valid_encryption(
        election_metadata.description_hash,
        context.elgamal_public_key,
        context.crypto_extended_base_hash,
    ):
        return encrypted_ballot
    else:
        return None  # log will have happened earlier



from dataclasses import dataclass, field
from typing import Dict, Optional, List

from .auxiliary import AuxiliaryDecrypt
from .data_store import DataStore
from .decryption import (
    compute_decryption_share,
    compute_compensated_decryption_share,
    reconstruct_missing_tally_decryption_shares,
)
from .decryption_share import TallyDecryptionShare, CompensatedTallyDecryptionShare
from .decrypt_with_shares import decrypt_tally
from .election import (
    CiphertextElectionContext,
    InternalElectionDescription,
)
from .election_polynomial import compute_lagrange_coefficient
from .group import ElementModP, ElementModQ
from .guardian import Guardian
from .key_ceremony import ElectionPublicKey
from .rsa import rsa_decrypt
from .tally import (
    CiphertextTally,
    PlaintextTally,
)
from .logs import log_info, log_warning
from .types import BALLOT_ID, GUARDIAN_ID

AVAILABLE_GUARDIAN_ID = GUARDIAN_ID
MISSING_GUARDIAN_ID = GUARDIAN_ID

GUARDIAN_PUBLIC_KEY = ElementModP


@dataclass
class DecryptionMediator:
    """
    The Decryption Mediator composes partial decryptions from each Guardian 
    to form a decrypted representation of an election tally
    """

    _metadata: InternalElectionDescription
    _encryption: CiphertextElectionContext

    _ciphertext_tally: CiphertextTally
    _plaintext_tally: Optional[PlaintextTally] = field(default=None)

    # Since spoiled ballots are decrypted, they are just a special case of a tally
    _plaintext_spoiled_ballots: Dict[BALLOT_ID, Optional[PlaintextTally]] = field(
        default_factory=lambda: {}
    )

    _available_guardians: Dict[AVAILABLE_GUARDIAN_ID, Guardian] = field(
        default_factory=lambda: {}
    )
    _missing_guardians: DataStore[MISSING_GUARDIAN_ID, ElectionPublicKey] = field(
        default_factory=lambda: DataStore()
    )

    _decryption_shares: Dict[AVAILABLE_GUARDIAN_ID, TallyDecryptionShare] = field(
        default_factory=lambda: {}
    )
    """
    A collection of Decryption Shares for each Available Guardian
    """

    _lagrange_coefficients: Dict[
        MISSING_GUARDIAN_ID, Dict[AVAILABLE_GUARDIAN_ID, ElementModQ]
    ] = field(default_factory=lambda: {})
    """
    A collection of lagrange coefficients `w_{i,j}` computed by available guardians for each missing guardian
    """

    _compensated_decryption_shares: Dict[
        MISSING_GUARDIAN_ID,
        Dict[AVAILABLE_GUARDIAN_ID, CompensatedTallyDecryptionShare],
    ] = field(default_factory=lambda: {})
    """
    A collection of Compensated Decryption Shares for each Available Guardian
    """

    def announce(self, guardian: Guardian) -> Optional[TallyDecryptionShare]:
        """
        Announce that a Guardian is present and participating in the decryption.  
        A Decryption Share will be generated for the Guardian

        :param guardian: The guardian who will participate in the decryption.
        :return: a `TallyDecryptionShare` for this `Guardian` or `None` if there is an error.
        """

        # Only allow a guardian to announce once
        if guardian.object_id in self._available_guardians:
            log_info(f"guardian {guardian.object_id} already announced")
            return self._decryption_shares[guardian.object_id]

        # Compute the Decryption Share for the guardian
        share = compute_decryption_share(
            guardian, self._ciphertext_tally, self._encryption
        )
        if share is None:
            log_warning(
                f"announce could not generate decryption share for {guardian.object_id}"
            )
            return None

        # Submit the share
        if self._submit_decryption_share(share):
            self._available_guardians[guardian.object_id] = guardian
        else:
            log_warning(
                f"announce could not submit decryption share for {guardian.object_id}"
            )
            return None

        # This guardian removes itself from the
        # missing list since it generated a valid share
        if guardian.object_id in self._missing_guardians.keys():
            self._missing_guardians.pop(guardian.object_id)

        # Check this guardian's collection of public keys
        # for other guardians that have not announced
        missing_guardians: Dict[MISSING_GUARDIAN_ID, ElectionPublicKey] = {
            guardian_id: public_key
            for guardian_id, public_key in guardian.guardian_election_public_keys()
            if guardian_id not in self._available_guardians
        }

        # Check that the public keys match for any missing guardians already reported
        # note this check naively assumes that the first guardian to annouce is telling the truth
        # but for this implementation it is simply a sanity check on the input data.
        # a consuming application should implement better validation of the guardian state
        # before announcing a guardian is available for decryption.
        for guardian_id, public_key in missing_guardians.items():
            if guardian_id in self._missing_guardians.keys():
                if self._missing_guardians.get(guardian_id) != public_key:
                    log_warning(
                        f"announce guardian: {guardian.object_id} expected public key mismatch for missing {guardian_id}"
                    )
                    return None
            else:
                self._missing_guardians.set(guardian_id, missing_guardians[guardian_id])

        return share

    def compensate(
        self, missing_guardian_id: str, decrypt: AuxiliaryDecrypt = rsa_decrypt
    ) -> Optional[List[CompensatedTallyDecryptionShare]]:
        """
        Compensate for a missing guardian by reconstructing the share using the available guardians.

        :param missing_guardian_id: the guardian that failed to `announce`.
        :return: a collection of `CompensatedTallyDecryptionShare` generated from all available guardians
                 or `None if there is an error
        """

        # Only allow a guardian to be compensated for once
        if missing_guardian_id in self._compensated_decryption_shares:
            log_warning(f"guardian {missing_guardian_id} already compensated")
            return list(
                self._compensated_decryption_shares[missing_guardian_id].values()
            )

        compensated_decryptions: List[CompensatedTallyDecryptionShare] = []
        lagrange_coefficients: Dict[AVAILABLE_GUARDIAN_ID, ElementModQ] = {}

        # Loop through each of the available guardians
        # and calculate a partial for the missing one
        for available_guardian in self._available_guardians.values():

            # Compute lagrange coefficients for each of the available guardians
            lagrange_coefficients[
                available_guardian.object_id
            ] = compute_lagrange_coefficient(
                available_guardian.sequence_order,
                *[
                    guardian.sequence_order
                    for guardian in self._available_guardians.values()
                    if guardian.object_id != available_guardian.object_id
                ],
            )

            # Compute the decryption shares
            share = compute_compensated_decryption_share(
                available_guardian,
                missing_guardian_id,
                self._ciphertext_tally,
                self._encryption,
                decrypt,
            )
            if share is None:
                log_warning(f"compensation failed for missing: {missing_guardian_id}")
                break
            else:
                compensated_decryptions.append(share)

        # Verify generated the correct number of partials
        if len(compensated_decryptions) != len(self._available_guardians):
            log_warning(
                f"compensate mismatch partial decryptions for missing guardian {missing_guardian_id}"
            )
            return None
        else:
            self._lagrange_coefficients[missing_guardian_id] = lagrange_coefficients
            self._submit_compensated_decryption_shares(compensated_decryptions)
            return compensated_decryptions

    def get_plaintext_tally(
        self, recompute: bool = False, decrypt: AuxiliaryDecrypt = rsa_decrypt
    ) -> Optional[PlaintextTally]:
        """
        Get the plaintext tally for the election by composing each Guardian's 
        decrypted representation of each selection into a decrypted representation

        :param recompute: Specify if the function should recompute the result, even if one already exists.
        :return: a `PlaintextTally` or `None`
        """

        if self._plaintext_tally is not None and not recompute:
            return self._plaintext_tally

        # Make sure a Quorum of Guardians have announced
        if len(self._available_guardians) < self._encryption.quorum:
            log_warning(
                "cannot get plaintext tally with less than quorum available guardians"
            )
            return None

        # If all Guardians are present decrypt the tally
        if len(self._available_guardians) == self._encryption.number_of_guardians:
            return decrypt_tally(
                self._ciphertext_tally, self._decryption_shares, self._encryption
            )

        # If missing guardians compensate for the missing guardians
        for missing in self._missing_guardians.keys():
            compensated_decryptions = self.compensate(missing, decrypt)
            if compensated_decryptions is None:
                log_warning(f"get plaintext tally failed compensating for {missing}")
                return None

        # Reconstruct the missing partial decryptions from the compensation shares
        missing_decryption_shares = reconstruct_missing_tally_decryption_shares(
            self._ciphertext_tally,
            self._missing_guardians,
            self._compensated_decryption_shares,
            self._lagrange_coefficients,
        )
        if missing_decryption_shares is None or len(missing_decryption_shares) != len(
            self._missing_guardians
        ):
            log_warning(f"get plaintext tally failed with missing decryption shares")
            return None

        merged_decryption_shares: Dict[str, TallyDecryptionShare] = {}

        for available, share in self._decryption_shares.items():
            merged_decryption_shares[available] = share

        for missing, share in missing_decryption_shares.items():
            merged_decryption_shares[missing] = share

        if len(merged_decryption_shares) != self._encryption.number_of_guardians:
            log_warning(f"get plaintext tally failed with share length mismatch")
            return None

        return decrypt_tally(
            self._ciphertext_tally, merged_decryption_shares, self._encryption
        )

    def _submit_decryption_share(self, share: TallyDecryptionShare) -> bool:
        """
        Submit the decryption share to be used in the decryption
        """
        if share.guardian_id in self._decryption_shares:
            log_warning(
                f"cannot submit for guardian {share.guardian_id} that already decrypted"
            )
            return False

        self._decryption_shares[share.guardian_id] = share
        return True

    def _submit_compensated_decryption_shares(
        self, shares: List[CompensatedTallyDecryptionShare]
    ) -> bool:
        """
        Submit compensated decryption shares to be used in the decryption
        """
        results = [self._submit_compensated_decryption_share(share) for share in shares]
        return all(results)

    def _submit_compensated_decryption_share(
        self, share: CompensatedTallyDecryptionShare
    ) -> bool:
        """
        Submit compensated decryption share to be used in the decryption
        """
        if (
            share.missing_guardian_id in self._compensated_decryption_shares
            and share.guardian_id
            in self._compensated_decryption_shares[share.missing_guardian_id]
        ):
            log_warning(
                f"cannot submit compensated share for guardian {share.guardian_id} on behalf of {share.missing_guardian_id} that already compensated"
            )
            return False

        if share.missing_guardian_id not in self._compensated_decryption_shares:
            self._compensated_decryption_shares[share.missing_guardian_id] = {}

        self._compensated_decryption_shares[share.missing_guardian_id][
            share.guardian_id
        ] = share

        return True



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from glob import glob
from os.path import basename, splitext, abspath, dirname, join

from setuptools import find_packages
from setuptools import setup

current_directory = abspath(dirname(__file__))
with open(join(current_directory, "README.md"), encoding="utf-8") as readme_file:
    LONG_DESCRIPTION = readme_file.read()

NAME = "electionguard"
VERSION = "1.1.4"
LICENSE = "MIT"
DESCRIPTION = "ElectionGuard: Support for e2e verified elections."
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"
AUTHOR = "Microsoft Corporation"
AUTHOR_EMAIL = "electionguard@microsoft.com"
URL = "https://github.com/microsoft/electionguard-python"
PROJECT_URLS = {
    "Documentation": "https://microsoft.github.io/electionguard-python",
    "Read the Docs": "https://electionguard-python.readthedocs.io",
    "Releases": "https://github.com/microsoft/electionguard-python/releases",
    "Milestones": "https://github.com/microsoft/electionguard-python/milestones",
    "Issue Tracker": "https://github.com/microsoft/electionguard-python/issues",
}
CLASSIFIERS = [
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: Unix",
    "Operating System :: POSIX",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Utilities",
]


setup(
    name=NAME,
    version=VERSION,
    license=LICENSE,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=CLASSIFIERS,
    project_urls=PROJECT_URLS,
    python_requires="~=3.8",
    install_requires=[
        "gmpy2>=2.0.8",
        "numpy>=1.18.2",
        "jsons>=1.1.2",
        "cryptography",
        "psutil>=5.7.2",
    ],
)



from dataclasses import dataclass

from .serializable import Serializable


@dataclass
class ElectionObjectBase(Serializable):
    """
    A base object to derive other election objects 
    that is both serializable and identifiable by object_id
    """

    object_id: str



from os import path
from typing import Iterable

from .ballot import PlaintextBallot, CiphertextBallot, CiphertextAcceptedBallot
from .guardian import Guardian
from .election import CiphertextElectionContext, ElectionConstants, ElectionDescription
from .encrypt import EncryptionDevice
from .key_ceremony import CoefficientValidationSet
from .tally import PlaintextTally, PublishedCiphertextTally
from .utils import make_directory

RESULTS_DIR = "results"
DEVICES_DIR = path.join(RESULTS_DIR, "devices")
COEFFICIENTS_DIR = path.join(RESULTS_DIR, "coefficients")
BALLOTS_DIR = path.join(RESULTS_DIR, "encrypted_ballots")
SPOILED_DIR = path.join(RESULTS_DIR, "spoiled_ballots")
PRIVATE_DIR = path.join(RESULTS_DIR, "private")
PLAINTEXT_BALLOTS_DIR = path.join(PRIVATE_DIR, "plaintext")
ENCRYPTED_BALLOTS_DIR = path.join(PRIVATE_DIR, "encrypted")

DESCRIPTION_FILE_NAME = "description"
CONTEXT_FILE_NAME = "context"
CONSTANTS_FILE_NAME = "constants"
ENCRYPTED_TALLY_FILE_NAME = "encrypted_tally"
TALLY_FILE_NAME = "tally"

DEVICE_PREFIX = "device_"
COEFFICIENT_PREFIX = "coefficient_validation_set_"
BALLOT_PREFIX = "ballot_"

PLAINTEXT_BALLOT_PREFIX = "plaintext_ballot_"
GUARDIAN_PREFIX = "guardian_"

# TODO #148 Revert PlaintextTally to PublishedPlaintextTally after moving spoiled info
def publish(
    description: ElectionDescription,
    context: CiphertextElectionContext,
    constants: ElectionConstants,
    devices: Iterable[EncryptionDevice],
    ciphertext_ballots: Iterable[CiphertextAcceptedBallot],
    spoiled_ballots: Iterable[CiphertextAcceptedBallot],
    ciphertext_tally: PublishedCiphertextTally,
    plaintext_tally: PlaintextTally,
    coefficient_validation_sets: Iterable[CoefficientValidationSet] = None,
    results_directory: str = RESULTS_DIR,
) -> None:
    """Publishes the election record as json"""

    make_directory(results_directory)
    description.to_json_file(DESCRIPTION_FILE_NAME, results_directory)
    context.to_json_file(CONTEXT_FILE_NAME, results_directory)
    constants.to_json_file(CONSTANTS_FILE_NAME, results_directory)

    make_directory(DEVICES_DIR)
    for device in devices:
        device_name = DEVICE_PREFIX + str(device.uuid)
        device.to_json_file(device_name, DEVICES_DIR)

    make_directory(COEFFICIENTS_DIR)
    if coefficient_validation_sets is not None:
        for coefficient_validation_set in coefficient_validation_sets:
            set_name = COEFFICIENT_PREFIX + coefficient_validation_set.owner_id
            coefficient_validation_set.to_json_file(set_name, COEFFICIENTS_DIR)

    make_directory(BALLOTS_DIR)
    for ballot in ciphertext_ballots:
        ballot_name = BALLOT_PREFIX + ballot.object_id
        ballot.to_json_file(ballot_name, BALLOTS_DIR)

    make_directory(SPOILED_DIR)
    for ballot in spoiled_ballots:
        ballot_name = BALLOT_PREFIX + ballot.object_id
        ballot.to_json_file(ballot_name, SPOILED_DIR)

    ciphertext_tally.to_json_file(ENCRYPTED_TALLY_FILE_NAME, results_directory)
    plaintext_tally.to_json_file(TALLY_FILE_NAME, results_directory)


def publish_private_data(
    plaintext_ballots: Iterable[PlaintextBallot],
    ciphertext_ballots: Iterable[CiphertextBallot],
    guardians: Iterable[Guardian],
    results_directory: str = RESULTS_DIR,
) -> None:
    """
    Publish the private data for an election.  
    Useful for generating sample data sets.  
    Do not use this in a production application.
    """

    make_directory(results_directory)
    make_directory(PRIVATE_DIR)

    for guardian in guardians:
        guardian_name = GUARDIAN_PREFIX + guardian.object_id
        guardian.to_json_file(guardian_name, PRIVATE_DIR, strip_privates=False)

    make_directory(PLAINTEXT_BALLOTS_DIR)
    for plaintext_ballot in plaintext_ballots:
        ballot_name = PLAINTEXT_BALLOT_PREFIX + plaintext_ballot.object_id
        plaintext_ballot.to_json_file(ballot_name, PLAINTEXT_BALLOTS_DIR)

    make_directory(ENCRYPTED_BALLOTS_DIR)
    for ciphertext_ballot in ciphertext_ballots:
        ballot_name = BALLOT_PREFIX + ciphertext_ballot.object_id
        ciphertext_ballot.to_json_file(ballot_name, ENCRYPTED_BALLOTS_DIR)
